# Copyright (c) 2022 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import abc
import gzip
import pathlib
import pickle  # nosec
import sys
import textwrap
import warnings
from multiprocessing import Process, Queue
from types import SimpleNamespace

import defusedxml.lxml
import repomd
import requests
from dogpile.cache.backends.null import NullBackend

from soufi import exceptions, finder

warnings.formatwarning = lambda msg, *x, **y: f"WARNING: {msg}\n"


class YumFinder(finder.SourceFinder, metaclass=abc.ABCMeta):
    """An abstract base class for making Yum-based finders.

    Subclasses of this should provide methods for setting up the repository
    search lists to pass to __init__.

    The lookup is a 2-stage process:
        1. Attempt to look up the SRPM directly from the sources repository.
           In these cases, we use the name only and ignore the version,
           since it is a very common practice for repositories to only
           publish repodata for the "current" version, while still
           providing older sources.
        2. Attempt to look up the name of the source RPM from the binary
           repository.  This will catch packages where the source and
           binary package names do not match.  The version is also ignored
           in this step, for the same reasons.  Then backtrack and attempt
           to lookup that package in the source repos.
    """

    def __init__(self, *args, source_repos=None, binary_repos=None, **kwargs):
        self.source_repos = source_repos
        self.binary_repos = binary_repos
        super().__init__(*args, **kwargs)
        if isinstance(self._cache.backend, NullBackend):
            warnings.warn(
                "Use of the Null cache with the DNF/Yum finder is highly "
                "ill-advised.  Please see the documentation.",
                stacklevel=1,
            )

    def generate_repos(self, repos, fallback):
        """Ensure a generator is always returned for repos.

        Either turn the init's value into a generator, or use the fallback
        methods that return the default ones as a generator.
        """
        if repos:
            yield from repos
        else:
            yield from fallback()

    def generate_source_repos(self):
        return self.generate_repos(self.source_repos, self.get_source_repos)

    def generate_binary_repos(self):
        return self.generate_repos(self.binary_repos, self.get_binary_repos)

    @abc.abstractmethod
    def get_source_repos(self):
        raise NotImplementedError  # pragma: nocover

    @abc.abstractmethod
    def get_binary_repos(self):
        raise NotImplementedError  # pragma: nocover

    def _find(self):
        source_url = self.get_source_url()
        return YumDiscoveredSource([source_url], timeout=self.timeout)

    def get_source_url(self):
        """Lookup the URL for the SRPM corresponding to the package name.

        :return: a URL of the location of an SRPM.
        :raises: exceptions.SourceNotFound if no SRPM could be found in any
            of the repos.
        :raises: exceptions.DownloadError on any failure downloading the
            repomd files.  The original exception that caused the failure
            may be inspected in the `__cause__` attribute.
        """
        # NOTE(nic): Try to find the package in the binary repos,
        #  then backtrack into the source repos with the name and version of
        #  the SRPM provided.  This is, in aggregate, faster than looking up
        #  the source first.
        try:
            source_name, source_ver = self._walk_binary_repos(self.name)
        except Exception as e:
            raise exceptions.DownloadError from e
        if source_name is None:
            raise exceptions.SourceNotFound

        try:
            url = self._walk_source_repos(source_name, source_ver)
        except Exception as e:
            raise exceptions.DownloadError from e
        if url is None:
            raise exceptions.SourceNotFound

        # If we have a URL, but it's no good, we don't have a URL.
        if not self.test_url(url):
            raise exceptions.SourceNotFound

        return url

    def _walk_source_repos(self, name, version=None):
        if version is None:
            version = self.version
        locations = set()
        for repo_url in self.generate_source_repos():
            baseurl, repo = self._cache.get_or_create(
                f"repo-{repo_url}",
                do_task,
                creator_args=([get_repomd, repo_url], {}),
            )
            for package in repo.get(name, []):
                # If the package version in the repomd is our version,
                # it's easy.  Note that we want to match epoch-full and
                # epoch-less version formats.
                if version in (package.evr, package.vr):
                    return str(baseurl + package.location)
                # Otherwise let's make it weird
                locations.add(
                    str(package.location.replace(package.vr, version))
                )

                # If we've made it here, things have gotten weird.  Replace the
                # version+release info in all unique package locations with our
                # version and see if they exist.  This should find superseded
                # packages that are present in the repo, but not in the repomd.
                for location in locations:
                    if self.test_url(baseurl + location):
                        return str(baseurl + location)
        return None

    def _walk_binary_repos(self, name):
        packages = set()
        for repo_url in self.generate_binary_repos():
            _, repo = self._cache.get_or_create(
                f"repo-{repo_url}",
                do_task,
                creator_args=([get_repomd, repo_url], {}),
            )
            for package in repo.get(name, []):
                # If we have a binary package matching our version, but
                # with a different name than the corresponding source
                # package, return the NVR fields
                if self.version in (package.evr, package.vr):
                    return self._nevra_or_none(package)
                # Otherwise let's make it weird
                packages.add(self._nevra_or_none(package))

        # If we've made it here, things have gotten weird; this should be
        # the case of a source RPM that produces binary RPMs with different
        # names *and* versions (the lvm2 package from Red Hat is a good example
        # of this).  In this case, we won't be able to make heads or tails of
        # the response, unless (and only unless) it contains a single package.
        try:
            [package] = packages
        except ValueError:
            return None, None
        return package

    def _nevra_or_none(self, package):
        if package.sourcerpm == '':
            # It's here, but has no sources defined!  Bummer...
            return None, None
        nevra = self._get_nevra(str(package.sourcerpm))
        return nevra['name'], f"{nevra['ver']}-{nevra['rel']}"

    # TODO(nic): throw this out and use hawkey/libdnf whenever that finally
    #  stabilizes.  See: https://github.com/juledwar/soufi/issues/13
    @staticmethod
    def _get_nevra(filename):
        """Split out the NEVRA fields from an RPM filename."""

        # It's easiest to do this by eating the filename backwards, dropping
        # offset pointers as we go
        if filename.endswith('.rpm'):
            filename = filename[:-4]
        arch_offset = filename.rfind('.')
        arch = filename[1 + arch_offset :]
        rel_offset = filename[:arch_offset].rfind('-')
        rel = filename[1 + rel_offset : arch_offset]
        ver_offset = filename[:rel_offset].rfind('-')
        ver = filename[1 + ver_offset : rel_offset]
        name = filename[:ver_offset]
        # Sometimes the epoch is before the name, sometimes it's before the
        # version.  Support both.
        if ':' in ver:
            epoch, ver = ver.split(':', maxsplit=1)
        elif ':' in name:
            epoch, name = name.split(':', maxsplit=1)
        else:
            epoch = ''
        return dict(name=name, ver=ver, rel=rel, epoch=epoch, arch=arch)


class YumDiscoveredSource(finder.DiscoveredSource):
    """A discovered Red Hat source package."""

    make_archive = finder.DiscoveredSource.remote_url_is_archive
    archive_extension = '.src.rpm'

    def populate_archive(self, *args, **kwargs):  # pragma: no cover
        # Src RPMs are already compressed archives, nothing to do.
        pass

    def __repr__(self):
        return self.urls[0]


# NOTE(nic): repomd objects require extra-special care and handling.  They
#  are thin wrappers around ElementTree objects, which means they use an
#  obnoxious amount of memory, are notoriously hostile to being pickled,
#  and are also averse to being automagically garbage-collected when they go
#  out of scope.  This puts them at odds with our caching strategy,
#  and makes them less-than-ideal for use inside long-running processes.
#
#  To deal with the dearth of effective memory management, we will run
#  all repomd lookups in subprocesses, that will return any/all "needles"
#  found in the "haystacks" we provide.  This will let the OS
#  efficiently reclaim all the pages used upon completion.
def do_task(target, *args):
    """Run the target callable in a subprocess and return its response."""
    queue = Queue()
    process = Process(target=target, args=(queue,) + args)
    try:
        process.start()
    except RuntimeError as e:
        if 'not using fork' in str(e):
            sys.exit(
                textwrap.dedent(
                    """
                FATAL: Running this finder directly from the global scope is
                not supported on this platform.  To use this finder, call it
                instead from the main module, e.g.:

                   if __name__ == '__main__':
                       soufi.finder.factory(*args, **kwargs).find()

                Aborting."""
                )
            )
        raise
    # We don't want to wait *forever*, but jobs can take several minutes to
    # complete, so wait a relatively long time
    response = queue.get(timeout=600)
    if process.is_alive():
        process.terminate()
    # re-raise exceptions thrown in child processes; this should keep them
    # from getting cached
    if response and isinstance(response[0], Exception):
        raise response[0]
    return response


# NOTE(nic): stolen almost verbatim from repomd.load, except this one:
#  - has timeouts
#  - uses requests instead of urllib to do the heavy lifting
#  - uses `lxml.parse` and file objects instead of `lxml.fromstring`
#  - returns a plain dict instead of a Repo object
def load_repomd(url):
    timeout = YumFinder.timeout
    baseurl = requests.utils.parse_url(url)
    path = pathlib.PurePosixPath(baseurl.path)

    # first we must get the repomd.xml file
    repomd_path = path / 'repodata' / 'repomd.xml'
    repomd_url = baseurl._replace(path=str(repomd_path))

    with requests.get(repomd_url, stream=True, timeout=timeout) as r:
        r.raw.decode_content = True
        repomd_xml = defusedxml.lxml.parse(r.raw)

    # determine the location of *primary.xml.gz from the repomd.xml
    primary_element = repomd_xml.find(
        'repo:data[@type="primary"]/repo:location', namespaces=repomd._ns
    )
    primary_path = path / primary_element.get('href')
    primary_url = baseurl._replace(path=str(primary_path))

    # download and consume *-primary.xml into a dict object for fast lookups.
    # We will use repomd.Package objects rather than reimplement them,
    # but only as an intermediate step (see `serialize_package`, below)
    repo = {}
    with requests.get(primary_url, stream=True, timeout=timeout) as r:
        r.raw.decode_content = True
        with gzip.GzipFile(fileobj=r.raw) as uncompressed:
            for element in defusedxml.lxml.parse(uncompressed).getroot():
                package = repomd.Package(element)
                repo.setdefault(package.name, [])
                repo[package.name].append(serialize_package(package))
    return repo


def get_repomd(queue, url):
    if not url.endswith('/'):
        url += '/'
    try:
        repo = load_repomd(url)
    except Exception as e:
        # Try and send exceptions back to the caller, just like anything
        # else.  It's up to the receiver to inspect and re-raise.  If the
        # exception cannot be serialized, send back a generic Exception.
        try:
            pickle.dumps(e)
        except Exception:
            e = Exception(
                f"Could not serialize {e.__class__.__name__}, "
                f"re-raising as plain Exception with msg: {str(e)}"
            )

        queue.put((e,), timeout=YumFinder.timeout)
        return
    queue.put((url, repo))


# NOTE(nic): repomd.Package object properties do XPath lookups into the
#  ElementTree and other similar tricks on the backend, which makes them
#  unsuitable for the IPC-based workflow we're trying to use, so we'll
#  convert them into simple objects with identical field names so that they
#  can be easily pickled and passed around.  The upside is that we only need
#  to carry over what we need to do package lookups, reducing the space
#  requirements by at least two orders of magnitude.  The downside is that we
#  only get what we've carried over, but in practice that's the opposite of
#  a problem.
def serialize_package(package):
    return SimpleNamespace(
        name=str(package.name),
        version=str(package.version),
        vr=str(package.vr),
        evr=str(package.evr),
        location=str(package.location),
        sourcerpm=str(package.sourcerpm),
    )



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import photon, yum
from soufi.testing import base


class BasePhotonTest(base.TestCase):
    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        if 'source_repos' not in kwargs:
            kwargs['source_repos'] = ['']
        if 'binary_repos' not in kwargs:
            kwargs['binary_repos'] = ['']
        return photon.PhotonFinder(name, version, SourceType.os, **kwargs)

    def make_response(self, data, code):
        fake_response = mock.MagicMock()
        fake_response.content = data
        fake_response.status_code = code
        return fake_response

    def make_href(self, text):
        return f'<a href="{text}">{text}</a>'

    def make_top_page_content(self, versions):
        links = [self.make_href(v) for v in versions]
        return "\n".join(links)


class TestPhotonFinder(BasePhotonTest):
    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, 'get_source_url').return_value = url
        disc_source = finder.find()
        self.assertIsInstance(disc_source, yum.YumDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test__get_source_repos(self):
        finder = self.make_finder()
        # Ensure irrelevant repo dirs are ignored with the 'bogus' ones.
        top_repos = ('3.0/', '4.0/')
        repos = (
            '1_srpms_x86_64/',
            'bogus/',
            '2_srpms_x86_64/',
            '3_srpms_x86_64/',
            'also_bogus/',
            '4_srpms_x86_64/',
        )
        top_data = self.make_top_page_content(top_repos)
        page1 = self.make_top_page_content(repos[:3])
        page2 = self.make_top_page_content(repos[3:])
        get = self.patch(requests, 'get')
        # Set up responses for the top index, and the packages indexes for each
        # item in repos.
        get.side_effect = (
            self.make_response(top_data, requests.codes.ok),
            self.make_response(page1, requests.codes.ok),
            self.make_response(page2, requests.codes.ok),
        )
        result = finder.get_source_repos()
        expected = [
            photon.PHOTON_PACKAGES + '/' + top_repos[1] + repos[0],
            photon.PHOTON_PACKAGES + '/' + top_repos[1] + repos[2],
            photon.PHOTON_PACKAGES + '/' + top_repos[0] + repos[3],
            photon.PHOTON_PACKAGES + '/' + top_repos[0] + repos[5],
        ]
        self.assertEqual(expected, result)
        # fmt: off
        get.assert_has_calls(
            [
                mock.call(photon.PHOTON_PACKAGES, timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/' + top_repos[1], timeout=30),  # noqa: E501
                mock.call(photon.PHOTON_PACKAGES + '/' + top_repos[0], timeout=30),  # noqa: E501
            ]
        )
        # fmt: on

    def test__get_source_repos_top_level_failure_throws_exception(self):
        finder = self.make_finder()
        get = self.patch(requests, 'get')
        get.return_value = self.make_response(b'', requests.codes.not_found)
        self.assertRaises(exceptions.DownloadError, finder.get_source_repos)

    def test__get_source_repos_subdir_failure_omits_subdirs(self):
        finder = self.make_finder()
        top_data = self.make_top_page_content(['1.0/', '2.0/'])
        data = self.make_top_page_content(['1_srpms_x86_64/'])
        get = self.patch(requests, 'get')
        # The first index page listed is actually no good
        get.side_effect = (
            self.make_response(top_data, requests.codes.ok),
            self.make_response(b'', requests.codes.not_found),
            self.make_response(data, requests.codes.ok),
        )
        # We should only have one source repo directory available
        result = finder.get_source_repos()
        expected = [f"{photon.PHOTON_PACKAGES}/1.0/1_srpms_x86_64/"]
        self.assertEqual(expected, result)
        get.assert_has_calls(
            [
                mock.call(photon.PHOTON_PACKAGES, timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/2.0/', timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/1.0/', timeout=30),
            ]
        )

    def test__get_binary_repos(self):
        finder = self.make_finder()
        top_repos = ('3.0/', '4.0/')
        repos = (
            '1_srpms_x86_64/',
            'bogus/',
            '2_packages_x86_64/',
            '3_updates_x86_64/',
            'also_bogus/',
            '4_srpms_x86_64/',
        )
        top_data = self.make_top_page_content(top_repos)
        page1 = self.make_top_page_content(repos[:3])
        page2 = self.make_top_page_content(repos[3:])
        self.patch(finder, 'test_url').return_value = True
        get = self.patch(requests, 'get')
        # Set up responses for the top index, and the packages indexes for each
        # item in repos.
        get.side_effect = (
            self.make_response(top_data, requests.codes.ok),
            self.make_response(page1, requests.codes.ok),
            self.make_response(page2, requests.codes.ok),
        )
        expected = [
            photon.PHOTON_PACKAGES + '/' + top_repos[1] + repos[2],
            photon.PHOTON_PACKAGES + '/' + top_repos[0] + repos[3],
        ]
        result = finder.get_binary_repos()
        self.assertEqual(expected, result)
        # fmt: off
        get.assert_has_calls(
            [
                mock.call(photon.PHOTON_PACKAGES, timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/' + top_repos[1], timeout=30),  # noqa: E501
                mock.call(photon.PHOTON_PACKAGES + '/' + top_repos[0], timeout=30),  # noqa: E501
            ]
        )
        # fmt: on

    def test__get_binary_repos_top_level_failure_throws_exception(self):
        finder = self.make_finder()
        get = self.patch(requests, 'get')
        get.return_value = self.make_response(b'', requests.codes.not_found)
        self.assertRaises(exceptions.DownloadError, finder.get_binary_repos)

    def test__get_binary_repos_subdir_failure_omits_subdirs(self):
        finder = self.make_finder()
        top_data = self.make_top_page_content(['6.0/', '9.0/'])
        data = self.make_top_page_content(['1_base_x86_64'])
        self.patch(finder, 'test_url').return_value = True
        get = self.patch(requests, 'get')
        # The first index page listed is actually no good
        get.side_effect = (
            self.make_response(top_data, requests.codes.ok),
            self.make_response(b'', requests.codes.not_found),
            self.make_response(data, requests.codes.ok),
        )
        # We should only have one source repo directory available
        result = finder.get_binary_repos()
        expected = [f"{photon.PHOTON_PACKAGES}/6.0/1_base_x86_64"]
        self.assertEqual(expected, result)
        get.assert_has_calls(
            [
                mock.call(photon.PHOTON_PACKAGES, timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/9.0/', timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/6.0/', timeout=30),
            ]
        )

    def test__get_binary_repos_subdir_failure_omits_empty_repo_dirs(self):
        finder = self.make_finder()
        top_data = self.make_top_page_content(['6.0/', '9.0/'])
        data = self.make_top_page_content(['1_base_x86_64'])
        # All repos in the top content are available
        get = self.patch(requests, 'get')
        get.side_effect = (
            self.make_response(top_data, requests.codes.ok),
            self.make_response(data, requests.codes.ok),
            self.make_response(data, requests.codes.ok),
        )
        # The first repo candidate contains no repo data
        self.patch(finder, 'test_url').side_effect = (False, True)
        # We should only have one source repo directory available
        result = finder.get_binary_repos()
        expected = [f"{photon.PHOTON_PACKAGES}/6.0/1_base_x86_64"]
        self.assertEqual(expected, result)
        get.assert_has_calls(
            [
                mock.call(photon.PHOTON_PACKAGES, timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/9.0/', timeout=30),
                mock.call(photon.PHOTON_PACKAGES + '/6.0/', timeout=30),
            ]
        )

    def test__walk_source_repos(self):
        name = self.factory.make_string('name')
        version = self.factory.make_string('ver')
        urls = [self.factory.make_url() for _ in range(10)]
        finder = self.make_finder(source_repos=urls)
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = (False, False, True)
        # Ensure that it doesn't run the entire list once one has been found
        self.assertEqual(
            f"{urls[2]}/{name}-{version}.src.rpm",
            finder._walk_source_repos(name, version),
        )
        test_url.assert_has_calls(
            [mock.call(f"{url}/{name}-{version}.src.rpm") for url in urls[:3]]
        )

    def test__walk_source_repos_no_match_returns_none(self):
        name = self.factory.make_string('name')
        version = self.factory.make_string('ver')
        urls = [self.factory.make_url() for _ in range(3)]
        finder = self.make_finder(source_repos=urls)
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = False
        self.assertIsNone(finder._walk_source_repos(name, version))
        test_url.assert_has_calls(
            [mock.call(f"{url}/{name}-{version}.src.rpm") for url in urls]
        )

    def test__walk_source_repos_one_arg_short_circuits(self):
        arg = self.factory.make_string()
        # Deliberately setup the mock wrong, this way if future code breaks
        # the short-circuit behaviour this will throw errors
        finder = self.make_finder(source_repos=[1, 2, 3])
        test_url = self.patch(finder, 'test_url')
        self.assertIsNone(finder._walk_source_repos(arg))
        test_url.assert_not_called()



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import pathlib
import tarfile
import tempfile
from unittest import mock

import fixtures
import testtools
from testtools.matchers import Equals
from testtools.matchers._basic import SameMembers

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import ubuntu
from soufi.testing import base


class TestUbuntuFinder(base.TestCase):
    def tearDown(self):
        super().tearDown()
        ubuntu.UbuntuFinder.get_archive.cache_clear()

    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return ubuntu.UbuntuFinder(name, version, SourceType.os, **kwargs)

    def test_init_auths_to_Launchpad_only_once(self):
        login = self.patch(ubuntu.Launchpad, 'login_anonymously')
        self.make_finder()
        self.make_finder()
        login.assert_called_once_with(
            "soufi", "production", mock.ANY, version="devel", timeout=30
        )

    def test_get_archive(self):
        lp = mock.MagicMock()
        distro = mock.MagicMock()
        lp.distributions.__getitem__.return_value = distro
        archive = mock.MagicMock()
        distro.main_archive = archive
        login = self.patch(ubuntu.Launchpad, 'login_anonymously')
        login.return_value = lp
        uf = self.make_finder()
        self.assertEqual(archive, uf.get_archive())

    def test_get_build(self):
        archive = mock.MagicMock()
        self.patch(ubuntu.UbuntuFinder, 'get_archive').return_value = archive
        getPublishedBinaries = mock.MagicMock()
        archive.getPublishedBinaries = getPublishedBinaries
        binary = mock.MagicMock()
        binary.build = mock.sentinel.BUILD
        getPublishedBinaries.return_value = [binary, mock.MagicMock()]
        uf = self.make_finder()
        build = uf.get_build()
        archive.getPublishedBinaries.assert_called_once_with(
            exact_match=True, binary_name=uf.name, version=uf.version
        )
        self.assertEqual(build, mock.sentinel.BUILD)

    def test_get_build_raises_for_no_build(self):
        archive = mock.MagicMock()
        self.patch(ubuntu.UbuntuFinder, 'get_archive').return_value = archive
        getPublishedBinaries = mock.MagicMock()
        archive.getPublishedBinaries = getPublishedBinaries
        getPublishedBinaries.return_value = []
        uf = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            uf.get_build()

    def test_get_source_from_build(self):
        archive = mock.MagicMock()
        self.patch(ubuntu.UbuntuFinder, 'get_archive').return_value = archive
        getPublishedSources = mock.MagicMock()
        archive.getPublishedSources = getPublishedSources
        source = mock.sentinel.SOURCE
        getPublishedSources.return_value = [source, mock.MagicMock()]
        build = mock.MagicMock()
        build.source_package_name = self.factory.make_string()
        build.source_package_version = self.factory.make_string()
        uf = self.make_finder()
        result = uf.get_source_from_build(build)
        getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name=build.source_package_name,
            version=build.source_package_version,
        )
        self.assertEqual(source, result)

    def test_find_returns_discovered_source(self):
        self.patch(ubuntu.UbuntuFinder, 'get_archive')
        self.patch(ubuntu.UbuntuFinder, 'get_build')
        source = mock.MagicMock()
        sourceFileUrls = mock.MagicMock()
        sourceFileUrls.return_value = [
            self.factory.make_url(),
            self.factory.make_url(),
        ]
        source.sourceFileUrls = sourceFileUrls
        self.patch(
            ubuntu.UbuntuFinder, 'get_source_from_build'
        ).return_value = source
        uf = self.make_finder()
        disc_source = uf.find()
        self.assertIsInstance(disc_source, ubuntu.UbuntuDiscoveredSource)
        self.assertThat(
            disc_source.urls, SameMembers(sourceFileUrls.return_value)
        )


class TestUbuntuDiscoveredSource(base.TestCase):
    def test_repr(self):
        urls = [self.factory.make_url() for _ in range(4)]
        uds = ubuntu.UbuntuDiscoveredSource(urls)
        expected = "\n".join(urls)
        self.assertEqual(expected, repr(uds))

    def test_populate_archive(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path

        # Make a file to put in the tar:
        content = self.factory.make_bytes('content')
        _, fake_file = tempfile.mkstemp(dir=tmpdir)
        with open(fake_file, 'wb') as fake_file_fd:
            fake_file_fd.write(content)

        # Patch out download_file to return the fake file:
        url = self.factory.make_url()
        uds = ubuntu.UbuntuDiscoveredSource([url])
        download_file = self.patch(uds, 'download_file')
        download_file.return_value = pathlib.Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            uds.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        expected_file_name = url.rsplit('/', 1)[-1]
        with tarfile.open(name=tar_file_name, mode='r') as tar:
            self.expectThat(tar.getnames(), SameMembers([expected_file_name]))
            self.expectThat(
                tar.extractfile(expected_file_name).read(), Equals(content)
            )



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest.mock import MagicMock

from soufi.finder import SourceType, factory
from soufi.finders import ubuntu
from soufi.testing import base


class TestFinderFactory(base.TestCase):
    def test_factory_loads_finder_classes(self):
        # Until the finder path is configurable, just check that the
        # Ubuntu finder is present.
        self.assertEqual(
            factory._finders[ubuntu.UbuntuFinder.distro], ubuntu.UbuntuFinder
        )

    def test_factory_passes_args_when_calling(self):
        name = self.factory.make_string()
        version = self.factory.make_string()
        finder_mock = MagicMock()
        factory._finders[ubuntu.UbuntuFinder.distro] = finder_mock
        factory(ubuntu.UbuntuFinder.distro, name, version, SourceType.os)
        finder_mock.assert_called_once_with(name, version, SourceType.os)

    def test_supported_types(self):
        expected = [
            'alpine',
            'centos',
            'debian',
            'gem',
            'go',
            'java',
            'npm',
            'photon',
            'python',
            'rhel',
            'ubuntu',
        ]
        self.assertListEqual(expected, factory.supported_types)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

# Bandit reports this as vulnerable but it's OK in lxml now,
# defusedxml's lxml support is deprecated as a result.
from lxml import html  # nosec

import soufi.exceptions
import soufi.finders.yum as yum_finder
from soufi import finder

PHOTON_PACKAGES = "https://packages.vmware.com/photon"


class PhotonFinder(yum_finder.YumFinder):
    """Find Photon source files.

    By default, Iterates over the index at https://packages.vmware.com/photon
    """

    distro = finder.Distro.photon.value

    def _get_dirs(self):
        content = self.get_url(PHOTON_PACKAGES).content
        tree = html.fromstring(content)
        retval = tree.xpath('//a/text()')
        return reversed([dir for dir in retval if dir[0].isdigit()])

    def _get_repos(self, xpath):
        dirs = []
        for release_dir in self._get_dirs():
            url = f"{PHOTON_PACKAGES}/{release_dir}"
            try:
                content = self.get_url(url).content
            except soufi.exceptions.DownloadError:
                continue
            tree = html.fromstring(content)
            # Ideally all the SRPM trees would have the exact same
            # packages in them, but their `aarch64` trees seem to be a
            # little light.  Prefer x86_64 to be safe
            dirs += [url + dir for dir in tree.xpath(xpath)]
        return dirs

    def get_source_repos(self):
        return self._get_repos(
            "//a[text()[contains(.,'srpms')][contains(.,'x86_64')]]/text()"
        )

    def get_binary_repos(self):
        """Retrieve a list of all repo URLs for binary RPM packages.

        Photon OS does not reliably publish repodata for all releases,
        so double-check all candidate repo dirs before using.
        """
        xpath = "//a[text()[not(contains(.,'srpms'))][contains(.,'x86_64')]]/text()"  # noqa: E501
        suffix = 'repodata/repomd.xml'
        return [r for r in self._get_repos(xpath) if self.test_url(r + suffix)]

    def _walk_source_repos(self, name, version=None):
        # Photon OS does not provide repomd.xml files for their source
        # repositories, so we need to override the wonderful source lookup
        # methods with...  this.

        # Short-circuit here to force a binary package lookup in the caller
        if version is None:
            return None

        # Re-assemble a source package name, and try to fetch it from all
        # the source repos.  This is startlingly effective.
        for repo in self.generate_source_repos():
            url = f"{repo.rstrip('/')}/{name}-{version}.src.rpm"
            if self.test_url(url):
                return url
        return None



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests
from testtools.matchers._basic import SameMembers

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import golang
from soufi.testing import base


class TestGolangFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        kwargs = dict(
            name=name,
            version=version,
            s_type=SourceType.go,
        )
        return golang.GolangFinder(**kwargs)

    def patch_requests(self, method_name, code):
        method = self.patch(requests, method_name)
        response = requests.Response()
        response.status_code = code
        response.close = mock.MagicMock()
        method.return_value = response
        return method

    def test_raises_when_module_not_found(self):
        finder = self.make_finder()
        self.patch_requests(
            'head',
            code=requests.codes.not_found,
        )
        self.assertRaises(exceptions.SourceNotFound, finder.find)

    def test_finds_module(self):
        finder = self.make_finder()
        head = self.patch_requests('head', code=requests.codes.ok)
        source = finder.find()
        expected = [
            f"{golang.PUBLIC_PROXY}{finder.name.lower()}"
            f"/@v/{finder.version}.zip"
        ]
        self.assertThat(source.urls, SameMembers(expected))
        head.assert_called_once_with(expected[0], timeout=finder.timeout)

    def test_retries_with_get_if_head_fails(self):
        finder = self.make_finder()
        head = self.patch_requests('head', code=requests.codes.not_allowed)
        get = self.patch_requests('get', code=requests.codes.ok)
        source = finder.find()
        expected = [
            f"{golang.PUBLIC_PROXY}{finder.name.lower()}"
            f"/@v/{finder.version}.zip"
        ]
        self.assertThat(source.urls, SameMembers(expected))
        head.assert_called_once_with(expected[0], timeout=finder.timeout)
        get.assert_called_once_with(
            expected[0], stream=True, timeout=finder.timeout
        )


class TestGolangDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return golang.GolangDiscoveredSource([url])

    def test_make_archive(self):
        gds = self.make_discovered_source()
        self.assertEqual(gds.make_archive, gds.remote_url_is_archive)

    def test_repr(self):
        url = self.factory.make_url()
        gds = self.make_discovered_source(url=url)
        self.assertEqual(url, repr(gds))



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import os
import tarfile
import tempfile
import urllib
import warnings
from hashlib import sha512
from io import BytesIO
from pathlib import Path

import fixtures
import testtools
from testtools.matchers import Equals
from testtools.matchers._basic import SameMembers

from soufi import exceptions, testing
from soufi.finder import SourceType
from soufi.finders import alpine
from soufi.testing import base


class TestAlpineFinder(base.TestCase):
    # This test suite uses static test data. This is against my normal
    # policies of using a factory where possible, but creating an aports
    # repo is just too complex. The data on disk contains a handful of
    # packages copied directly from the 3.14-stable release, which cover all
    # the tests that need to happen.

    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        self.aports_dir = Path(testing.__path__[0]) / 'data' / 'aports'
        kwargs = dict(
            name=name,
            version=version,
            s_type=SourceType.os,
            aports_dir=self.aports_dir,
        )
        return alpine.AlpineFinder(**kwargs)

    def test_raises_when_package_not_found(self):
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.find()

    def test_raises_when_package_exists_but_no_version_match(self):
        finder = self.make_finder('dbus', '9999-0')
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.find()

    def test_finds_simple_package(self):
        # The simple case where a package name matches a directory in the
        # aports repo.
        finder = self.make_finder('dbus', '1.12.20-2')
        source = finder.find()
        file_prefix = f"file://{self.aports_dir}/main/dbus"
        expected = [
            "https://dbus.freedesktop.org/releases/dbus/dbus-1.12.20.tar.gz",
            f"{file_prefix}/0001-_dbus_generate_random_bytes-use-getrandom-2.patch",  # noqa: E501
            f"{file_prefix}/avoid-opendir-between-fork-exec.patch",
            f"{file_prefix}/dbus.initd",
        ]
        self.assertThat(source.urls, SameMembers(expected))

    def test_finds_simple_package_with_r_prefix_on_release(self):
        # When passing versions as version-rN, ensure it still matches.
        finder = self.make_finder('dbus', '1.12.20-r2')
        source = finder.find()
        file_prefix = f"file://{self.aports_dir}/main/dbus"
        expected = [
            "https://dbus.freedesktop.org/releases/dbus/dbus-1.12.20.tar.gz",
            f"{file_prefix}/0001-_dbus_generate_random_bytes-use-getrandom-2.patch",  # noqa: E501
            f"{file_prefix}/avoid-opendir-between-fork-exec.patch",
            f"{file_prefix}/dbus.initd",
        ]
        self.assertThat(source.urls, SameMembers(expected))

    def test_finds_subpackage(self):
        # Complex case where the target to find is a subpackage of a source.
        finder = self.make_finder('dbus-dev', '1.12.20-2')
        source = finder.find()
        file_prefix = f"file://{self.aports_dir}/main/dbus"
        expected = [
            "https://dbus.freedesktop.org/releases/dbus/dbus-1.12.20.tar.gz",
            f"{file_prefix}/0001-_dbus_generate_random_bytes-use-getrandom-2.patch",  # noqa: E501
            f"{file_prefix}/avoid-opendir-between-fork-exec.patch",
            f"{file_prefix}/dbus.initd",
        ]
        self.assertThat(source.urls, SameMembers(expected))

    def test_finds_subpackage_in_provides(self):
        # Similar to above, except the subpackage is in the "provides".
        finder = self.make_finder('xproto', '2021.4-0')
        source = finder.find()
        expected = [
            "https://xorg.freedesktop.org/archive/individual/proto/xorgproto-2021.4.tar.bz2"  # noqa: E501
        ]
        self.assertThat(source.urls, SameMembers(expected))

    def test_finds_subpackage_with_function(self):
        # Ensure we can still match package:_function in the subpackages
        finder = self.make_finder('libcrypto1.1', '1.1.1k-0')
        source = finder.find()
        file_prefix = f"file://{self.aports_dir}/main/openssl"
        expected = [
            'https://www.openssl.org/source/openssl-1.1.1k.tar.gz',
            f'{file_prefix}/man-section.patch',
            f'{file_prefix}/ppc64.patch',
        ]
        self.assertThat(source.urls, SameMembers(expected))

    def test_ignores_virtual_packages(self):
        # The lua5.2 5.2.4-7 package has a provides="lua" line, with no
        # version.  Check that this package is ignored.
        finder = self.make_finder(name="lua", version="5.2.4-7")
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.find()

    def test_catches_badly_formed_APKBUILD(self):
        name = self.factory.make_string()
        tmpdir = self.useFixture(fixtures.TempDir()).path
        apkbuild_dir = Path(tmpdir) / 'main' / name
        os.makedirs(apkbuild_dir)
        content = self.factory.make_bytes('content')
        with open(apkbuild_dir / 'APKBUILD', 'wb') as fake_file_fd:
            fake_file_fd.write(content)

        finder = alpine.AlpineFinder(
            name=name, version='bar', aports_dir=tmpdir, s_type=SourceType.os
        )
        with testtools.ExpectedException(exceptions.DownloadError):
            finder.find()

    def test_processes_ftp_sources(self):
        # The gdbm package has a single ftp source in it.
        finder = self.make_finder('gdbm', '1.13-1')
        source = finder.find()
        self.assertThat(
            source.urls,
            SameMembers(['ftp://ftp.nluug.nl/pub/gnu/gdbm/gdbm-1.13.tar.gz']),
        )


class TestAlpineDiscoveredSource(base.TestCase):
    def make_discovered_source(self, urls=None):
        if urls is None:
            urls = [self.factory.make_url()]
        return alpine.AlpineDiscoveredSource(urls)

    def make_file_with_content(self, target_dir, content: bytes = None):
        if content is None:
            content = self.factory.make_bytes('content')
        _, fake_file = tempfile.mkstemp(dir=target_dir)
        with open(fake_file, 'wb') as fake_file_fd:
            fake_file_fd.write(content)
        return fake_file

    def assert_tarfile_name_and_content(
        self, tar_file_name, expected_file_name, content
    ):
        # Test that the tar contains the file and the content.
        with tarfile.open(name=tar_file_name, mode='r') as tar:
            self.expectThat(tar.getnames(), SameMembers([expected_file_name]))
            self.expectThat(
                tar.extractfile(expected_file_name).read(), Equals(content)
            )

    def test_repr(self):
        urls = [self.factory.make_url() for _ in range(3)]
        ads = alpine.AlpineDiscoveredSource(urls)
        expected = "\n".join(urls)
        self.assertEqual(expected, repr(ads))

    def test_populate_archive(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        content = self.factory.make_bytes('content')
        fake_file = self.make_file_with_content(tmpdir, content)

        # Patch out download_file to return the fake file:
        url = self.factory.make_url()
        sha512sum = {os.path.basename(url): sha512(content).hexdigest()}
        ads = alpine.AlpineDiscoveredSource([url], sha512sums=sha512sum)
        download_file = self.patch(ads, 'download_file')
        download_file.return_value = Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            ads.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        expected_file_name = url.rsplit('/', 1)[-1]
        self.assert_tarfile_name_and_content(
            tar_file_name, expected_file_name, content
        )

    def test_populate_archive_warns_on_missing_checksum(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        content = self.factory.make_bytes('content')
        fake_file = self.make_file_with_content(tmpdir, content)
        warn = self.patch(warnings, 'warn')

        # Patch out download_file to return the fake file:
        url = self.factory.make_url()
        pathname = os.path.basename(url)
        ads = alpine.AlpineDiscoveredSource([url])
        download_file = self.patch(ads, 'download_file')
        download_file.return_value = Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            ads.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        expected_file_name = url.rsplit('/', 1)[-1]
        self.assert_tarfile_name_and_content(
            tar_file_name, expected_file_name, content
        )
        warn.assert_called_once_with(
            f'No checksum for source file {pathname}, cannot verify',
            stacklevel=1,
        )

    def test_populate_archive_raises_on_mismatched_checksum(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        content = self.factory.make_bytes('content')
        fake_file = self.make_file_with_content(tmpdir, content)

        # Patch out download_file to return the fake file:
        url = self.factory.make_url()
        pathname = os.path.basename(url)
        sha512sum = {pathname: 'invalid'}
        ads = alpine.AlpineDiscoveredSource([url], sha512sums=sha512sum)
        download_file = self.patch(ads, 'download_file')
        download_file.return_value = Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            exc = self.assertRaises(
                exceptions.DownloadError, ads.populate_archive, tmpdir, tar
            )
            self.assertIsNone(exc.status_code)

    def test_populate_archive_with_file_scheme(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        content = self.factory.make_bytes('content')
        fake_file = self.make_file_with_content(tmpdir, content)

        url = "file://" + str(fake_file)
        sha512sum = {os.path.basename(url): sha512(content).hexdigest()}
        ads = alpine.AlpineDiscoveredSource([url], sha512sums=sha512sum)
        download_file = self.patch(ads, 'download_file')
        download_file.return_value = Path(fake_file)

        # Make the tar file and populate it:
        tmpdir2 = self.useFixture(fixtures.TempDir()).path
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir2)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            ads.populate_archive(tmpdir2, tar)

        # Test that the tar contains the file and the content.
        expected_file_name = url.rsplit('/', 1)[-1]
        self.assert_tarfile_name_and_content(
            tar_file_name, expected_file_name, content
        )

    def test_populate_archive_with_custom_filename_prefix(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        content = self.factory.make_bytes('content')
        fake_file = self.make_file_with_content(tmpdir, content)

        # Patch out download_file to return the fake file:
        custom_name = self.factory.make_string("name")
        url = f"{custom_name}::{self.factory.make_url()}"
        sha512sum = {custom_name: sha512(content).hexdigest()}
        ads = alpine.AlpineDiscoveredSource([url], sha512sums=sha512sum)
        download_file = self.patch(ads, 'download_file')
        download_file.return_value = Path(fake_file)

        # Make the tar file and populate it:
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            ads.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        self.assert_tarfile_name_and_content(
            tar_file_name, custom_name, content
        )

    def test_download_ftp_file(self):
        tmpdir = self.useFixture(fixtures.TempDir()).path
        url = self.factory.make_url(scheme='ftp')
        content = self.factory.make_bytes('content')
        response = BytesIO(content)
        urlopen = self.patch(urllib.request, 'urlopen')
        urlopen.return_value = response
        sha512sum = {os.path.basename(url): sha512(content).hexdigest()}

        ads = alpine.AlpineDiscoveredSource(urls=[url], sha512sums=sha512sum)
        _, tar_file_name = tempfile.mkstemp(dir=tmpdir)
        with tarfile.open(name=tar_file_name, mode='w') as tar:
            ads.populate_archive(tmpdir, tar)

        # Test that the tar contains the file and the content.
        expected_file_name = url.rsplit('/', 1)[-1]
        self.assert_tarfile_name_and_content(
            tar_file_name, expected_file_name, content
        )



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

from unittest import mock

import requests
import testtools

from soufi import exceptions
from soufi.finder import SourceType
from soufi.finders import npm
from soufi.testing import base


class TestNPMFinder(base.TestCase):
    def make_finder(self, name=None, version=None):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        return npm.NPMFinder(name, version, SourceType.npm)

    def make_response(self, data, code):
        fake_response = mock.MagicMock()
        fake_response.json.return_value = data
        fake_response.status_code = code
        return fake_response

    def test_get_source_url(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        data = dict(dist=dict(tarball=url))

        get = self.patch(requests, 'get')
        get.return_value = self.make_response(data, requests.codes.ok)
        found_url = finder.get_source_url()
        self.assertEqual(found_url, url)
        get.assert_called_once_with(
            f"https://registry.npmjs.org/{finder.name}/{finder.version}",
            timeout=finder.timeout,
        )

    def test_get_source_info_raises_when_response_fails(self):
        get = self.patch(requests, 'get')
        get.return_value = self.make_response([], requests.codes.not_found)
        finder = self.make_finder()
        with testtools.ExpectedException(exceptions.SourceNotFound):
            finder.get_source_url()

    def test_find(self):
        url = self.factory.make_url()
        finder = self.make_finder()
        self.patch(finder, 'get_source_url').return_value = url

        disc_source = finder.find()
        self.assertIsInstance(disc_source, npm.NPMDiscoveredSource)
        self.assertEqual([url], disc_source.urls)


class TestNPMDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return npm.NPMDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        nds = self.make_discovered_source(url)
        self.assertEqual(url, repr(nds))

    def test_make_archive(self):
        nds = self.make_discovered_source()
        self.assertEqual(nds.make_archive, nds.remote_url_is_archive)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

"""All exceptions raised by Soufi."""


class SourceNotFound(Exception):
    """Raised when source cannot be located."""

    pass


class DownloadError(Exception):
    """Raised when source cannot be downloaded.

    This is most commonly raised due to HTTPError exceptions, so this will
    also surface the status code from any chained HTTP responses, as a
    convenience.
    """

    @property
    def status_code(self):
        for arg in self.args:
            try:
                return arg.response.status_code
            except AttributeError:
                continue
        return None



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import itertools
from unittest import mock

import requests

from soufi.finder import SourceType
from soufi.finders import centos, yum
from soufi.testing import base


class BaseCentosTest(base.TestCase):
    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        if 'source_repos' not in kwargs:
            kwargs['source_repos'] = ['']
        if 'binary_repos' not in kwargs:
            kwargs['binary_repos'] = ['']
        return centos.CentosFinder(name, version, SourceType.os, **kwargs)

    def make_response(self, data, code):
        fake_response = mock.MagicMock()
        fake_response.content = data
        fake_response.status_code = code
        return fake_response

    def make_href(self, text):
        return f'<a href="{text}">{text}</a>'

    def make_td(self, href):
        return f'<td class="indexcolname">{href}</td>'

    def make_top_page_content(self, versions):
        links = [self.make_td(self.make_href(v)) for v in versions]
        return "\n".join(links)


class TestCentosFinder(BaseCentosTest):
    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, 'get_source_url').return_value = url
        disc_source = finder.find()
        self.assertIsInstance(disc_source, yum.YumDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test__get_dirs(self):
        finder = self.make_finder()
        top_repos = ('1.0.123', '2.1.3456', 'bogus', '3.7.89', '3')
        top_data = self.make_top_page_content(top_repos)
        get = self.patch(requests, 'get')
        get.return_value = self.make_response(top_data, requests.codes.ok)
        result = list(finder._get_dirs())
        # Ensure that only the items we're interested in come back
        self.assertEqual(['3.7.89', '2.1.3456', '1.0.123'], result)
        get.assert_called_once_with(centos.VAULT, timeout=30)

    def test__get_source_repos(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_source_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/Source/"
            for dir in dirs
            for subdir in subdirs
        ]
        self.assertEqual(expected, result)

    def test__get_source_repos_optimal(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs, optimal_repos=True)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_source_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/Source/"
            for dir in dirs
            for subdir in (centos.DEFAULT_SEARCH + centos.OPTIMAL_SEARCH)
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/x86_64/os/"
            for dir in dirs
            for subdir in subdirs
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_old_style(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/x86_64/"
            for dir in dirs
            for subdir in subdirs
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_using_mirror(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.MIRROR}{dir}/{subdir}/x86_64/os/"
            for dir in dirs
            for subdir in subdirs
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_old_style_using_mirror(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, False, False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.MIRROR}{dir}/{subdir}/x86_64/"
            for dir in dirs
            for subdir in subdirs
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_optimal(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs, optimal_repos=True)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.return_value = True
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/x86_64/os/"
            for dir in dirs
            for subdir in (centos.DEFAULT_SEARCH + centos.OPTIMAL_SEARCH)
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_optimal_old_style(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs, optimal_repos=True)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.VAULT}{dir}/{subdir}/x86_64/"
            for dir in dirs
            for subdir in (centos.DEFAULT_SEARCH + centos.OPTIMAL_SEARCH)
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_optimal_using_mirror(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs, optimal_repos=True)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.MIRROR}{dir}/{subdir}/x86_64/os/"
            for dir in dirs
            for subdir in (centos.DEFAULT_SEARCH + centos.OPTIMAL_SEARCH)
        ]
        self.assertEqual(expected, result)

    def test__get_binary_repos_optimal_old_style_using_mirror(self):
        subdirs = [self.factory.make_string() for _ in range(3)]
        finder = self.make_finder(repos=subdirs, optimal_repos=True)
        dirs = [self.factory.make_string() for _ in range(3)]
        get_dirs = self.patch(finder, '_get_dirs')
        get_dirs.return_value = dirs
        test_url = self.patch(finder, 'test_url')
        test_url.side_effect = itertools.cycle((False, False, False, True))
        result = list(finder.get_binary_repos())
        expected = [
            f"{centos.MIRROR}{dir}/{subdir}/x86_64/"
            for dir in dirs
            for subdir in (centos.DEFAULT_SEARCH + centos.OPTIMAL_SEARCH)
        ]
        self.assertEqual(expected, result)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.

import logging
import tempfile
from subprocess import run as run_cmd  # nosec

import testtools

from soufi import finder
from soufi.finder import SourceType

# Setup a module-level cache dict so that it can be warmed up and re-used
# between tests
FUNCTEST_CACHE = dict()

# These tests will use the `memory_pickle` cache backend, which
# serializes the cache payload before storing it; this should surface
# any obvious problems with the other serializing cache backends
# without forcing them as a dependency.


class FunctionalFinderTests(testtools.TestCase):
    def setUp(self):
        # Enable dogpile.cache logging so we can eyeball the cache
        # population steps
        logging.basicConfig()
        logging.getLogger("dogpile.cache").setLevel(logging.DEBUG)
        super().setUp()


class FunctionalPhotonTests(FunctionalFinderTests):
    def test_find_superseded_package(self):
        photon = finder.factory(
            'photon',
            name='curl-libs',
            version='7.75.0-2.ph4',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://packages.vmware.com/photon/4.0/photon_srpms_4.0_x86_64/curl-7.75.0-2.ph4.src.rpm'  # noqa: E501
        result = photon.find()
        self.assertEqual([url], result.urls)


class FunctionalCentOSTests(FunctionalFinderTests):
    def test_find_binary_from_source(self):
        # Test finding sources for binary packages with entirely different
        # names/versions.
        centos = finder.factory(
            'centos',
            name='device-mapper-libs',
            version='1.02.175-5.el8',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://vault.centos.org/centos/8.4.2105/BaseOS/Source/SPackages/lvm2-2.03.11-5.el8.src.rpm'  # noqa: E501
        result = centos.find()
        self.assertEqual([url], result.urls)

    def test_find_binary_from_source_epoch(self):
        # Ibid, but with the epoch included with the version
        centos = finder.factory(
            'centos',
            name='device-mapper-libs',
            version='8:1.02.175-5.el8',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://vault.centos.org/centos/8.4.2105/BaseOS/Source/SPackages/lvm2-2.03.11-5.el8.src.rpm'  # noqa: E501
        result = centos.find()
        self.assertEqual([url], result.urls)

    def test_find_binary_from_source_epoch_no_vault(self):
        # Ibid, but with a binary package not tracked in vault.  This test
        # will need updating if/when 7.9.2009 eventually gets vaulted.
        centos = finder.factory(
            'centos',
            name='bind-license',
            version='32:9.11.4-26.P2.el7',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://vault.centos.org/centos/7.9.2009/os/Source/SPackages/bind-9.11.4-26.P2.el7.src.rpm'  # noqa: E501
        result = centos.find()
        self.assertEqual([url], result.urls)


class FunctionalRedhatTests(FunctionalFinderTests):
    def test_find_binary_from_source(self):
        # Test finding sources for binary packages with entirely different
        # names/versions.
        rhel = finder.factory(
            'rhel',
            name='vim-minimal',
            version='7.4.629-8.el7_9',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://cdn-ubi.redhat.com/content/public/ubi/dist/ubi/server/7/7Server/x86_64/source/SRPMS/Packages/v/vim-7.4.629-8.el7_9.src.rpm'  # noqa: E501
        result = rhel.find()
        self.assertEqual([url], result.urls)

    def test_find_binary_from_source_epoch(self):
        # Ibid, but with the epoch included with the version
        rhel = finder.factory(
            'rhel',
            name='vim-minimal',
            version='2:7.4.629-8.el7_9',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://cdn-ubi.redhat.com/content/public/ubi/dist/ubi/server/7/7Server/x86_64/source/SRPMS/Packages/v/vim-7.4.629-8.el7_9.src.rpm'  # noqa: E501
        result = rhel.find()
        self.assertEqual([url], result.urls)


class FunctionalAlpineTests(FunctionalFinderTests):
    def test_find_package(self):
        cmd = 'git clone --depth 1 --branch v3.13.5 git://git.alpinelinux.org/aports'.split()  # noqa: E501
        with tempfile.TemporaryDirectory('.aports') as aports_dir:
            output = run_cmd(cmd + [aports_dir], capture_output=True)  # nosec
            msg = "\n".join(
                [output.stderr.decode('utf-8'), output.stdout.decode('utf-8')]
            )
            if 'fatal' in msg:
                self.fail(f"git clone failed: {cmd} {aports_dir} {msg}")
            alpine = finder.factory(
                'alpine',
                name='zlib',
                version='1.2.11-r3',
                aports_dir=aports_dir,
                s_type=SourceType.os,
                cache_backend='dogpile.cache.memory_pickle',
                cache_args=dict(cache_dict=FUNCTEST_CACHE),
            )
            url = 'https://zlib.net/zlib-1.2.11.tar.gz'
            result = alpine.find()
            self.assertEqual([url], result.urls)


class FunctionalUbuntuTests(FunctionalFinderTests):
    def test_find_package(self):
        ubuntu = finder.factory(
            'ubuntu',
            name='openssh-client',
            version='1:7.6p1-4ubuntu0.3',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        urls = (
            'https://launchpad.net/ubuntu/+archive/primary/+sourcefiles/openssh/1:7.6p1-4ubuntu0.3/openssh_7.6p1-4ubuntu0.3.debian.tar.xz',  # noqa: E501
            'https://launchpad.net/ubuntu/+archive/primary/+sourcefiles/openssh/1:7.6p1-4ubuntu0.3/openssh_7.6p1-4ubuntu0.3.dsc',  # noqa: E501
            'https://launchpad.net/ubuntu/+archive/primary/+sourcefiles/openssh/1:7.6p1-4ubuntu0.3/openssh_7.6p1.orig.tar.gz',  # noqa: E501
            'https://launchpad.net/ubuntu/+archive/primary/+sourcefiles/openssh/1:7.6p1-4ubuntu0.3/openssh_7.6p1.orig.tar.gz.asc',  # noqa: E501
        )
        result = ubuntu.find()
        self.assertEqual(urls, result.urls)


class FunctionalDebianTests(FunctionalFinderTests):
    def test_find_package(self):
        debian = finder.factory(
            'debian',
            name='liblz4-1',
            version='1.8.3-1',
            s_type=SourceType.os,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        names = (
            'lz4_1.8.3.orig.tar.gz',
            'lz4_1.8.3-1.dsc',
            'lz4_1.8.3-1.debian.tar.xz',
        )
        urls = (
            'https://snapshot.debian.org/file/070867abcd93a7245b80ec6fc2ced27c6b8e3e0c',  # noqa: E501
            'https://snapshot.debian.org/file/c8bea6da056133de5da5ddedd0b1a65a19241f52',  # noqa: E501
            'https://snapshot.debian.org/file/85218375b0ba0ff7f989f85909d5560efdd2592d',  # noqa: E501
        )
        result = debian.find()
        self.assertEqual(names, result.names)
        self.assertEqual(urls, result.urls)


class FunctionalGemTests(FunctionalFinderTests):
    def test_find_package(self):
        gem = finder.factory(
            'gem',
            name='msgpack',
            version='1.3.1',
            s_type=SourceType.gem,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://rubygems.org/downloads/msgpack-1.3.1.gem'
        result = gem.find()
        self.assertEqual([url], result.urls)


class FunctionalGolangTests(FunctionalFinderTests):
    def test_find_package(self):
        golang = finder.factory(
            'go',
            name='github.com/beevik/ntp',
            version='v0.3.0',
            s_type=SourceType.go,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://proxy.golang.org/github.com/beevik/ntp/@v/v0.3.0.zip'
        result = golang.find()
        self.assertEqual([url], result.urls)


class FunctionalJavaTests(FunctionalFinderTests):
    def test_find_package(self):
        java = finder.factory(
            'java',
            name='log4j-core',
            version='2.9.1',  # vulnerable to log4shell
            s_type=SourceType.java,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://repo1.maven.org/maven2/org%2Fapache%2Flogging%2Flog4j%2Flog4j-core%2F2.9.1%2Flog4j-core-2.9.1-sources.jar'  # noqa: E501
        result = java.find()
        self.assertEqual([url], result.urls)


class FunctionalNPMTests(FunctionalFinderTests):
    def test_find_package(self):
        npm = finder.factory(
            'npm',
            name='react',
            version='16.14.0',
            s_type=SourceType.npm,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://registry.npmjs.org/react/-/react-16.14.0.tgz'
        result = npm.find()
        self.assertEqual([url], result.urls)


class FunctionalPythonTests(FunctionalFinderTests):
    def test_find_package(self):
        python = finder.factory(
            'python',
            name='SQLAlchemy',
            version='1.4.31',  # current version ;-)
            s_type=SourceType.python,
            cache_backend='dogpile.cache.memory_pickle',
            cache_args=dict(cache_dict=FUNCTEST_CACHE),
        )
        url = 'https://files.pythonhosted.org/packages/0f/80/d8883f12689a55e333d221bb9a56c727e976f5a8e9dc862efeac9f40d296/SQLAlchemy-1.4.31.tar.gz'  # noqa: E501
        result = python.find()
        self.assertEqual([url], result.urls)



# Copyright (c) 2021 Cisco Systems, Inc. and its affiliates
# All rights reserved.
import io
import string
from itertools import repeat
from unittest import mock

import repomd
import requests
from testtools.matchers import Equals, SameMembers

import soufi.exceptions
from soufi.finder import SourceType
from soufi.finders import yum
from soufi.testing import base


class YumFinderImpl(yum.YumFinder):
    """An implementation for testing the YumFimder ABC."""

    distro = 'yum'

    def get_source_repos(self):
        return ()

    def get_binary_repos(self):
        return ()


class BaseYumTest(base.TestCase):
    class FakePackage:
        vr = None
        evr = None
        location = None
        sourcerpm = None

        def __init__(self, vr=None, evr=None, location=None, sourcerpm=None):
            if vr is None:
                vr = BaseYumTest.factory.make_semver()
            if evr is None:
                evr = BaseYumTest.factory.make_semver()
            if location is None:
                location = BaseYumTest.factory.make_string()
            if sourcerpm is None:
                sourcerpm = BaseYumTest.factory.make_string()

            self.vr = vr
            self.evr = evr
            self.location = location
            self.sourcerpm = sourcerpm

    def make_finder(self, name=None, version=None, **kwargs):
        if name is None:
            name = self.factory.make_string('name')
        if version is None:
            version = self.factory.make_string('version')
        if 'source_repos' not in kwargs:
            kwargs['source_repos'] = []
        if 'binary_repos' not in kwargs:
            kwargs['binary_repos'] = []
        return YumFinderImpl(name, version, SourceType.os, **kwargs)

    def make_response(self, data, code):
        fake_response = mock.MagicMock()
        fake_response.content = data
        fake_response.status_code = code
        return fake_response

    def make_package(self, n=None, e=None, v=None, r=None, a=None, epoch=None):
        # `epoch` is `name` or `ver`, to denote where to inject it
        random_packagename = map(
            self.factory.random_choice,
            repeat(string.ascii_letters + string.digits + "_-"),
        )
        if n is None:
            n = self.factory.make_string(charset=random_packagename)
        if e is None:
            e = self.factory.randint(0, 10)
        if v is None:
            v = self.factory.make_semver()
        if r is None:
            r = self.factory.randint(0, 100)
        if a is None:
            a = self.factory.make_string()
        if epoch == 'ver':
            v = f"{e}:{v}"
        elif epoch == 'name':
            n = f"{e}:{n}"

        # Package names always end in `.rpm` for testing
        return f"{n}-{v}-{r}.{a}.rpm"


class TestYumFinder(BaseYumTest):
    def test_find(self):
        finder = self.make_finder()
        url = self.factory.make_url()
        self.patch(finder, 'get_source_url').return_value = url
        disc_source = finder.find()
        self.assertIsInstance(disc_source, yum.YumDiscoveredSource)
        self.assertEqual([url], disc_source.urls)

    def test_generate_repos_with_plain_list(self):
        fallback = mock.MagicMock()
        finder = self.make_finder()
        repos = [self.factory.make_string() for _ in range(0, 3)]
        result = finder.generate_repos(repos, fallback)
        self.assertThat(result, SameMembers(repos))
        fallback.assert_not_called()

    def test_generate_repos_with_fallback(self):
        fallback = mock.MagicMock()
        repos_fallback = [self.factory.make_string() for _ in range(0, 3)]
        fallback.return_value = repos_fallback
        finder = self.make_finder()
        result = finder.generate_repos(None, fallback)
        self.assertThat(list(result), SameMembers(repos_fallback))

    def test_get_source_url(self):
        name = self.factory.make_string()
        ver = self.factory.make_semver(extra=self.factory.make_string("-"))
        url = self.factory.make_url()
        finder = self.make_finder(name=name, version=ver)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_src.return_value = url
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.return_value = name, ver
        self.patch(finder, 'test_url').return_value = True
        self.assertEqual(url, finder.get_source_url())
        walk_binary.assert_called_once_with(name)
        walk_src.assert_called_once_with(name, ver)

    def test_get_source_url_raises_on_no_binary_package(self):
        name = self.factory.make_string()
        finder = self.make_finder(name=name)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.return_value = (None, None)
        self.assertRaises(
            soufi.exceptions.SourceNotFound, finder.get_source_url
        )
        walk_binary.assert_called_once_with(name)
        walk_src.assert_not_called()

    def test_get_source_url_raises_on_binary_repomd_download_error(self):
        name = self.factory.make_string()
        finder = self.make_finder(name=name)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.side_effect = Exception()
        self.assertRaises(
            soufi.exceptions.DownloadError, finder.get_source_url
        )
        walk_binary.assert_called_once_with(name)
        walk_src.assert_not_called()

    def test_get_source_url_raises_on_no_source(self):
        name = self.factory.make_string()
        finder = self.make_finder(name=name)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_src.return_value = None
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.return_value = (name, None)
        self.assertRaises(
            soufi.exceptions.SourceNotFound, finder.get_source_url
        )
        walk_binary.assert_called_once_with(name)
        walk_src.assert_called_once_with(name, None)

    def test_get_source_url_raises_on_invalid_source(self):
        name = self.factory.make_string()
        url = self.factory.make_url()
        finder = self.make_finder(name=name)
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.return_value = (name, None)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_src.return_value = url
        self.patch(finder, 'test_url').return_value = False
        self.assertRaises(
            soufi.exceptions.SourceNotFound, finder.get_source_url
        )
        walk_binary.assert_called_once_with(name)
        walk_src.assert_called_once_with(name, None)

    def test_get_source_url_raises_on_source_repomd_download_error(self):
        name = self.factory.make_string()
        finder = self.make_finder(name=name)
        walk_src = self.patch(finder, '_walk_source_repos')
        walk_src.side_effect = Exception()
        walk_binary = self.patch(finder, '_walk_binary_repos')
        walk_binary.return_value = (name, None)
        self.assertRaises(
            soufi.exceptions.DownloadError, finder.get_source_url
        )
        walk_binary.assert_called_once_with(name)
        walk_src.assert_called_once_with(name, None)

    def test__walk_source_repos(self):
        baseurl = self.factory.make_url()
        src = [self.factory.make_url(), self.factory.make_url()]
        bin = [self.factory.make_url(), self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        package = self.FakePackage(vr=finder.version)
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (baseurl, {finder.name: [package]})
        self.patch(finder, 'test_url').return_value = False
        url = finder._walk_source_repos(finder.name)
        self.assertEqual(baseurl + package.location, url)

    def test__walk_source_repos_different_version_hit(self):
        baseurl = self.factory.make_url()
        src = [self.factory.make_url()]
        bin = [self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        package = self.FakePackage()
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (baseurl, {finder.name: [package]})
        self.patch(finder, 'test_url').return_value = True
        url = finder._walk_source_repos(finder.name)
        self.assertEqual(baseurl + package.location, url)

    def test__walk_source_repos_different_version_miss(self):
        baseurl = self.factory.make_url()
        src = [self.factory.make_url()]
        bin = [self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        package = self.FakePackage()
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (baseurl, {finder.name: [package]})
        self.patch(finder, 'test_url').return_value = False
        url = finder._walk_source_repos(finder.name)
        self.assertIsNone(url)

    def test__walk_binary_repos(self):
        src = [self.factory.make_url(), self.factory.make_url()]
        bin = [self.factory.make_url(), self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        n = self.factory.make_string()
        v = self.factory.make_semver()
        r = self.factory.randint(0, 100)
        srcrpm = self.make_package(n=n, v=v, r=r)
        package = self.FakePackage(vr=finder.version, sourcerpm=srcrpm)
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (None, {finder.name: [package]})
        name, version = finder._walk_binary_repos(finder.name)
        self.expectThat(name, Equals(n))
        self.expectThat(version, Equals(f"{v}-{r}"))

    def test__walk_binary_repos_no_sourcerpm(self):
        src = [self.factory.make_url()]
        bin = [self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        package = self.FakePackage(vr=finder.version, sourcerpm='')
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (None, {finder.name: [package]})
        name, version = finder._walk_binary_repos(finder.name)
        self.assertIsNone(name)
        self.assertIsNone(version)

    def test__walk_binary_repos_different_name_multiple_versions(self):
        src = [self.factory.make_url()]
        bin = [self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        package1 = self.FakePackage()
        package2 = self.FakePackage()
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (None, {finder.name: [package1, package2]})
        name, version = finder._walk_binary_repos(finder.name)
        self.assertIsNone(name)
        self.assertIsNone(version)

    def test__walk_binary_repos_different_name_different_version(self):
        src = [self.factory.make_url()]
        bin = [self.factory.make_url()]
        finder = self.make_finder(source_repos=src, binary_repos=bin)
        n = self.factory.make_string()
        v = self.factory.make_semver()
        r = self.factory.randint(0, 100)
        srcrpm = self.make_package(n=n, v=v, r=r)
        package = self.FakePackage(sourcerpm=srcrpm)
        do_task = self.patch(yum, 'do_task')
        do_task.return_value = (None, {finder.name: [package]})
        name, version = finder._walk_binary_repos(finder.name)
        self.expectThat(name, Equals(n))
        self.expectThat(version, Equals(f"{v}-{r}"))


class TestYumFinderClassHelpers(BaseYumTest):
    def test__nevra_or_none_returns_nevra(self):
        name = self.factory.make_string()
        ver = self.factory.make_semver()
        rel = self.factory.randint(0, 100)
        pkg = mock.MagicMock()
        pkg.sourcerpm = self.make_package(n=name, v=ver, r=rel)
        finder = self.make_finder()
        self.assertEqual((name, f"{ver}-{rel}"), finder._nevra_or_none(pkg))

    def test__nevra_or_none_returns_none(self):
        package = mock.MagicMock()
        package.sourcerpm = ''
        finder = self.make_finder()
        self.assertEqual((None, None), finder._nevra_or_none(package))

    def test__get_nevra(self):
        filename = self.make_package()
        finder = self.make_finder()
        nevra = finder._get_nevra(filename)
        reassembled = "{name}-{ver}-{rel}.{arch}.rpm".format(**nevra)
        self.assertEqual(filename, reassembled)

    def test__get_nevra_epoch_in_ver(self):
        e = self.factory.randint(0, 10)
        filename = self.make_package(e=e, epoch='ver')
        finder = self.make_finder()
        nevra = finder._get_nevra(filename)
        reassembled = "{name}-{epoch}:{ver}-{rel}.{arch}.rpm".format(**nevra)
        self.assertEqual(filename, reassembled)

    def test__get_nevra_epoch_in_name(self):
        e = self.factory.randint(0, 10)
        filename = self.make_package(e=e, epoch='name')
        finder = self.make_finder()
        nevra = finder._get_nevra(filename)
        reassembled = "{epoch}:{name}-{ver}-{rel}.{arch}.rpm".format(**nevra)
        self.assertEqual(filename, reassembled)

    def test__test_url_true(self):
        url = self.factory.make_url()
        fake_req = self.patch(requests, 'head')
        fake_req.return_value = self.make_response('', requests.codes.ok)
        finder = self.make_finder()
        self.assertTrue(finder.test_url(url))

    def test__test_url_false(self):
        url = self.factory.make_url()
        fake_req = self.patch(requests, 'head')
        fake_req.return_value = self.make_response('', requests.codes.teapot)
        finder = self.make_finder()
        self.assertFalse(finder.test_url(url))

    def test__test_url_timeout(self):
        url = self.factory.make_url()
        self.patch(requests, 'head').side_effect = requests.exceptions.Timeout
        finder = self.make_finder()
        self.assertFalse(finder.test_url(url))


class TestYumFinderHelpers(BaseYumTest):
    def setUp(self):
        self.queue = mock.MagicMock()
        super().setUp()

    def test_get_repomd(self):
        # Mock up a successful repomd fetch
        url = self.factory.make_url()
        self.patch(requests, 'get')
        lxml = self.patch(repomd.defusedxml.lxml, 'parse')
        package = mock.MagicMock()
        lxml.return_value.getroot.return_value = [package]

        yum.get_repomd(self.queue, url)
        # Mocking out all the various and sundry Package properties would be
        # tedious, and we don't really care about anything past the package
        # name anyhow
        self.queue.put.assert_called_once_with(
            (url + "/", {package.findtext.return_value: [mock.ANY]})
        )

    def test_get_repomd_http_error(self):
        # Mock up a failure to fetch the repomd
        url = self.factory.make_url()
        load = self.patch(requests, 'get')
        load.side_effect = requests.exceptions.HTTPError()
        lxml = self.patch(repomd.defusedxml.lxml, 'parse')

        # Ensure that get_repomd won't fill the cache with garbage
        yum.get_repomd(self.queue, url)
        lxml.assert_not_called()
        self.queue.put.assert_called_once_with((load.side_effect,), timeout=30)

    def test_get_repomd_unserializable_http_error(self):
        # Ibid, but initializing the exception with a live file pointer will
        # make it refuse to serialize
        url = self.factory.make_url()
        fp = io.BufferedReader(io.StringIO())
        load = self.patch(requests, 'get')
        load.side_effect = requests.exceptions.RequestException(fp)
        lxml = self.patch(repomd.defusedxml.lxml, 'parse')

        # Ensure that we get a re-raised plain Exception
        yum.get_repomd(self.queue, url)
        lxml.assert_not_called()
        self.queue.put.assert_called_once_with((mock.ANY,), timeout=30)
        self.assertIn(
            're-raising as plain Exception', str(self.queue.put.call_args)
        )

    def test_do_task(self):
        # Mock up a process that does not exit upon return
        data = self.factory.make_string('response')
        queue = self.patch(yum, 'Queue')
        queue.return_value.get.return_value = data
        process = self.patch(yum, 'Process')
        process.return_value.is_alive.return_value = True

        # Ensure that the process gets shot in the head
        response = yum.do_task('a', 'b', 'c')
        self.assertEqual(data, response)
        process.return_value.terminate.assert_called_once_with()

    def test_do_task_empty_response(self):
        # Mock up a process that yields an "empty-but-successful" response
        queue = self.patch(yum, 'Queue')
        queue.return_value.get.return_value = []
        process = self.patch(yum, 'Process')
        process.return_value.is_alive.return_value = False

        # Ensure that the process gets shot in the head
        response = yum.do_task('d', 'e', 'f')
        self.assertEqual([], response)
        process.return_value.terminate.assert_not_called()

    def test_do_task_reraises_exceptions(self):
        # Actually run a job via do_task and ensure that the resulting
        # exception is intact
        data = self.factory.make_string('response')
        err = self.assertRaises(RuntimeError, yum.do_task, kaboom, data)
        self.assertEqual(data, str(err))

    def test_do_task_handles_spawn_errors_on_silly_platforms(self):
        # This simulates calling `process.start()` from the global scope on
        # platforms that do not support such things.  The default traceback
        # is not intrinsically helpful, so test that we kick back a more
        # useful error message.  See issue #31.
        process = self.patch(yum, 'Process')
        process.return_value.start.side_effect = RuntimeError(
            'Simulating a platform that is not using fork to start children'
        )
        err = self.assertRaises(SystemExit, yum.do_task, None)
        self.assertIn('FATAL: ', str(err))

    def test_do_task_leaves_other_spawn_errors_alone(self):
        # As per the above, other RuntimeError exceptions should pass through
        process = self.patch(yum, 'Process')
        process.return_value.start.side_effect = RuntimeError
        self.assertRaises(RuntimeError, yum.do_task, None)


# A simple subprocess function that throws a test exception.  Used by
# TestYumFinderHelpers.test_do_task_reraises_exceptions
def kaboom(queue, data):
    try:
        raise RuntimeError(data)
    except Exception as e:
        queue.put((e,))
    # This should never get called, but just in case...
    queue.put([])


class TestYumDiscoveredSource(base.TestCase):
    def make_discovered_source(self, url=None):
        if url is None:
            url = self.factory.make_url()
        return yum.YumDiscoveredSource([url])

    def test_repr(self):
        url = self.factory.make_url()
        yds = self.make_discovered_source(url)
        self.assertEqual(url, repr(yds))

    def test_make_archive(self):
        yds = self.make_discovered_source()
        self.assertEqual(yds.make_archive, yds.remote_url_is_archive)
